/*
 * Copyright (C) 1997 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file afh_common.c Common audio format handler functions. */

#include <sys/mman.h> /* mmap */
#include <sys/types.h>
#include <regex.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"

typedef void afh_init_func(struct audio_format_handler *);
/* It does not hurt to declare init functions which are not available. */
extern afh_init_func mp3_init, ogg_init, aac_afh_init, wma_afh_init,
	spx_afh_init, flac_afh_init, opus_afh_init;

/** The list of all status items */
const char *status_item_list[] = {STATUS_ITEM_ARRAY};

/**
 * The list of supported audio formats.
 *
 * We always define the full array of audio formats even if some audio formats
 * were not compiled in. This is because for each audio file the number of its
 * audio format is stored in the database. We don't want that numbers to become
 * stale just because the user installed a new version of paraslash that
 * supports a different set of audio formats.
 *
 * It can still be easily detected whether an audio format is compiled in by
 * checking if the init function pointer is not \p NULL.
 */
static struct audio_format_handler afl[] = {
	{
		.name = "mp3",
		.init = mp3_init,
	},
	{
		.name = "ogg",
#if defined(HAVE_OGG) && defined(HAVE_VORBIS)
		.init = ogg_init,
#endif
	},
	{
		.name = "aac",
#if defined(HAVE_MP4V2)
		.init = aac_afh_init,
#endif
	},
	{
		.name = "wma",
		.init = wma_afh_init,
	},
	{
		.name = "spx",
#if defined(HAVE_OGG) && defined(HAVE_SPEEX)
		.init = spx_afh_init,
#endif
	},
	{
		.name = "flac",
#if defined(HAVE_OGG) && defined(HAVE_FLAC)
		.init = flac_afh_init,
#endif
	},
	{
		.name = "opus",
#if defined(HAVE_OGG) && defined(HAVE_OPUS)
		.init = opus_afh_init,
#endif
	},
	{
		.name = NULL,
	}
};

static inline int next_audio_format(int format)
{
	for (;;) {
		if (!afl[format].name)
			return format;
		format++;
		if (afl[format].init)
			return format;
	}

}

/** Iterate over each supported audio format. */
#define FOR_EACH_AUDIO_FORMAT(i) for (i = 0; afl[i].name; i = next_audio_format(i))

/**
 * Call the init function of each supported audio format handler.
 */
void afh_init(void)
{
	int i;

	PARA_NOTICE_LOG("supported audio formats: %s\n", AUDIO_FORMAT_HANDLERS);
	FOR_EACH_AUDIO_FORMAT(i) {
		PARA_INFO_LOG("initializing %s handler\n",
			audio_format_name(i));
		afl[i].init(&afl[i]);
	}
}

/**
 * Guess the audio format judging from filename.
 *
 * \param name The filename.
 *
 * \return This function returns \p -E_AUDIO_FORMAT if it has no idea what kind
 * of audio file this might be. Otherwise the (non-negative) number of the
 * audio format is returned.
 */
int guess_audio_format(const char *name)
{
	int i,j, len = strlen(name);

	FOR_EACH_AUDIO_FORMAT(i) {
		for (j = 0; afl[i].suffixes[j]; j++) {
			const char *p = afl[i].suffixes[j];
			int plen = strlen(p);
			if (len < plen + 1)
				continue;
			if (name[len - plen - 1] != '.')
				continue;
			if (strcasecmp(name + len - plen, p))
				continue;
//			PARA_DEBUG_LOG("might be %s\n", audio_format_name(i));
			return i;
		}
	}
	return -E_AUDIO_FORMAT;
}

/**
 * Get the name of the given audio format.
 *
 * \param i The audio format number.
 *
 * \return This returns a pointer to statically allocated memory so it
 * must not be freed by the caller.
 */
const char *audio_format_name(int i)
{
	if (i < 0 || i >= ARRAY_SIZE(afl) - 1)
		return "???";
	return afl[i].name;
}

static int get_file_info(int format, const char *path, char *data,
		size_t size, int fd, struct afh_info *afhi)
{
	int ret;
	const char *fmt = audio_format_name(format);

	memset(afhi, 0, sizeof(*afhi));
	ret = afl[format].get_file_info(data, size, fd, afhi);
	if (ret < 0) {
		PARA_WARNING_LOG("%s: %s format not detected: %s\n",
			path, fmt, para_strerror(-ret));
		return ret;
	}
	PARA_NOTICE_LOG("%s: detected %s format\n", path, fmt);
	return format;
}

/**
 * Call get_file_info() to obtain an afhi structure.
 *
 * \param path The full path of the audio file.
 * \param data Pointer to the contents of the (mapped) file.
 * \param size The file size in bytes.
 * \param fd The open file descriptor.
 * \param afhi Result pointer.
 *
 * \return The number of the audio format on success, \p -E_AUDIO_FORMAT if no
 * compiled in audio format handler is able to handler the file.
 *
 * This function tries to find an audio format handler that can interpret the
 * file given by \a data and \a size.
 *
 * It first tries to determine the audio format from the filename given by \a
 * path. If this doesn't work, all other audio format handlers are tried until
 * one is found that can handle the file.
 */
int compute_afhi(const char *path, char *data, size_t size, int fd,
		struct afh_info *afhi)
{
	int ret, i, format;

	format = guess_audio_format(path);
	if (format >= 0) {
		ret = get_file_info(format, path, data, size, fd, afhi);
		if (ret >= 0)
			goto success;
	}
	FOR_EACH_AUDIO_FORMAT(i) {
		if (i == format) /* we already tried this one to no avail */
			continue;
		ret = get_file_info(i, path, data, size, fd, afhi);
		if (ret >= 0)
			goto success;
	}
	return -E_AUDIO_FORMAT;
success:
	if (!afhi->techinfo)
		afhi->techinfo = para_strdup(NULL);
	if (!afhi->tags.artist)
		afhi->tags.artist = para_strdup(NULL);
	if (!afhi->tags.title)
		afhi->tags.title = para_strdup(NULL);
	if (!afhi->tags.year)
		afhi->tags.year = para_strdup(NULL);
	if (!afhi->tags.album)
		afhi->tags.album = para_strdup(NULL);
	if (!afhi->tags.comment)
		afhi->tags.comment = para_strdup(NULL);
	PARA_DEBUG_LOG("techinfo: %s\n", afhi->techinfo);
	PARA_DEBUG_LOG("artist: %s\n", afhi->tags.artist);
	PARA_DEBUG_LOG("title: %s\n", afhi->tags.title);
	PARA_DEBUG_LOG("year: %s\n", afhi->tags.year);
	PARA_DEBUG_LOG("album: %s\n", afhi->tags.album);
	PARA_DEBUG_LOG("comment: %s\n", afhi->tags.comment);
	return ret;
}

/**
 * Deallocate the contents of an afh_info structure.
 *
 * \param afhi The structure to clear.
 *
 * This only frees the memory the various pointer fields of \a afhi point to.
 * It does *not* free \a afhi itself.
 */
void clear_afhi(struct afh_info *afhi)
{
	if (!afhi)
		return;
	free(afhi->chunk_table);
	free(afhi->techinfo);
	free(afhi->tags.artist);
	free(afhi->tags.title);
	free(afhi->tags.year);
	free(afhi->tags.album);
	free(afhi->tags.comment);
}

static inline size_t get_chunk_len(long unsigned chunk_num,
		const struct afh_info *afhi)
{
	return afhi->chunk_table[chunk_num + 1] - afhi->chunk_table[chunk_num];
}

/**
 * Get one chunk of audio data.
 *
 * \param chunk_num The number of the chunk to get.
 * \param afhi Describes the audio file.
 * \param map The memory mapped audio file.
 * \param buf Result pointer.
 * \param len The length of the chunk in bytes.
 *
 * Upon return, \a buf will point so memory inside \a map. The returned buffer
 * must therefore not be freed by the caller.
 */
void afh_get_chunk(long unsigned chunk_num, struct afh_info *afhi,
		void *map, const char **buf, size_t *len)
{
	size_t pos = afhi->chunk_table[chunk_num];
	*buf = map + pos;
	*len = get_chunk_len(chunk_num, afhi);
}

/**
 * Find a suitable start chunk.
 *
 * \param approx_chunk_num Upper bound for the chunk number to return.
 * \param afhi Needed for the chunk table.
 *
 * \return The first non-empty chunk <= \a approx_chunk_num.
 *
 * \sa \ref afh_get_chunk().
 */
int32_t afh_get_start_chunk(int32_t approx_chunk_num,
		const struct afh_info *afhi)
{
	int32_t k;

	for (k = PARA_MAX(0, approx_chunk_num); k >= 0; k--)
		if (get_chunk_len(k, afhi) > 0)
			break;
	return k;
}

/**
 * Get the header of an audio file.
 *
 * \param afhi The audio file handler data describing the file.
 * \param audio_format_id Determines the audio format handler.
 * \param map The data of the audio file.
 * \param mapsize The amount of bytes of the mmapped audio file.
 * \param buf The length of the header is stored here.
 * \param len Points to a buffer containing the header on return.
 *
 * This function sets \a buf to \p NULL and \a len to zero if \a map or \a
 * afhi is \p NULL, or if the current audio format does not need special
 * header treatment.
 *
 * Otherwise, it is checked whether the audio format handler given by
 * \a audio_format_id defines a ->get_header() method. If it does, this
 * method is called to obtain the header. If ->get_header() is \p NULL,
 * a reference to the first chunk of the audio file is returned.
 *
 * Once the header is no longer needed, the caller must call \ref
 * afh_free_header() to free the resources allocated by this function.
 */
void afh_get_header(struct afh_info *afhi, uint8_t audio_format_id,
		void *map, size_t mapsize, char **buf, size_t *len)
{
	struct audio_format_handler *afh = afl + audio_format_id;

	if (!map || !afhi || !afhi->header_len) {
		*buf = NULL;
		*len = 0;
		return;
	}
	if (!afh->get_header) {
		*len = afhi->header_len;
		*buf = map;
		return;
	}
	afh->get_header(map, mapsize, buf, len);
}

/**
 * Deallocate any resources obtained from afh_get_header().
 *
 * \param header_buf Pointer obtained via afh_get_header().
 * \param audio_format_id Determines the audio format handler.
 */
void afh_free_header(char *header_buf, uint8_t audio_format_id)
{
	struct audio_format_handler *afh = afl + audio_format_id;

	if (afh->get_header)
		free(header_buf);
}

/**
 * Pretty-print the contents of a struct afh_info into a buffer.
 *
 * \param audio_format_num The audio format number.
 * \param afhi Pointer to the structure that contains the information.
 * \param result Pretty-printed ahfi is here after the call.
 *
 * The \a result buffer is dynamically allocated and should be freed by the
 * caller.
 *
 * \return The number of bytes. This function never fails.
 */
unsigned afh_get_afhi_txt(int audio_format_num, struct afh_info *afhi, char **result)
{
	return xasprintf(result, "%s: %dkbit/s\n" /* bitrate */
		"%s: %s\n" /* format */
		"%s: %dHz\n" /* frequency */
		"%s: %d\n" /* channels */
		"%s: %" PRIu32 "\n" /* seconds total */
		"%s: %lu: %lu\n" /* chunk time */
		"%s: %" PRIu32 "\n" /* num chunks */
		"%s: %s\n" /* techinfo */
		"%s: %s\n" /* artist */
		"%s: %s\n" /* title */
		"%s: %s\n" /* year */
		"%s: %s\n" /* album */
		"%s: %s\n", /* comment */
		status_item_list[SI_BITRATE], afhi->bitrate,
		status_item_list[SI_FORMAT], audio_format_name(audio_format_num),
		status_item_list[SI_FREQUENCY], afhi->frequency,
		status_item_list[SI_CHANNELS], afhi->channels,
		status_item_list[SI_SECONDS_TOTAL], afhi->seconds_total,
		status_item_list[SI_CHUNK_TIME], (long unsigned)afhi->chunk_tv.tv_sec,
			(long unsigned)afhi->chunk_tv.tv_usec,
		status_item_list[SI_NUM_CHUNKS], afhi->chunks_total,
		status_item_list[SI_TECHINFO], afhi->techinfo? afhi->techinfo : "",
		status_item_list[SI_ARTIST], afhi->tags.artist? afhi->tags.artist : "",
		status_item_list[SI_TITLE], afhi->tags.title? afhi->tags.title : "",
		status_item_list[SI_YEAR], afhi->tags.year? afhi->tags.year : "",
		status_item_list[SI_ALBUM], afhi->tags.album? afhi->tags.album : "",
		status_item_list[SI_COMMENT], afhi->tags.comment? afhi->tags.comment : ""
	);
}

/**
 * Create a copy of the given file with altered meta tags.
 *
 * \param audio_format_id Specifies the audio format.
 * \param map The (read-only) memory map of the input file.
 * \param mapsize The size of the input file in bytes.
 * \param tags The new tags.
 * \param output_fd Altered file is created using this file descriptor.
 * \param filename The name of the temporary output file.
 *
 * This calls the ->rewrite_tags method of the audio format handler associated
 * with \a audio_format_id to create a copy of the memory-mapped file given
 * by \a map and \a mapsize, but with altered tags according to \a tags. If
 * the audio format handler for \a audio_format_id lacks this optional method,
 * the function returns (the paraslash error code of) \p ENOTSUP.
 *
 * \return Standard.
 */
int afh_rewrite_tags(int audio_format_id, void *map, size_t mapsize,
		struct taginfo *tags, int output_fd, const char *filename)
{
	struct audio_format_handler *afh = afl + audio_format_id;

	if (!afh->rewrite_tags)
		return -ERRNO_TO_PARA_ERROR(ENOTSUP);
	return afh->rewrite_tags(map, mapsize, tags, output_fd, filename);
}
