/*
 * Copyright (C) 2004 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file ogg_afh.c Audio format handler for ogg/vorbis files. */

#include <vorbis/codec.h>
#include <regex.h>

#include "para.h"
#include "afh.h"
#include "error.h"
#include "string.h"
#include "ogg_afh_common.h"

struct private_vorbis_data {
	vorbis_info vi;
	vorbis_comment vc;
};

/*
 * Vorbis uses three header packets, all of which are required: the
 * identification header, the comments header, and the setup header.
 *
 * The identification header identifies the bitstream as Vorbis. It contains
 * the Vorbis version and simple audio characteristics of the stream such as
 * sample rate and number of channels.
 *
 * The comment header includes user text comments (tags) and a vendor string
 * for the application/library that produced the bitstream.
 *
 * The setup header includes extensive CODEC setup information as well as the
 * complete VQ and Huffman codebooks needed for decoding.
 */
static int vorbis_packet_callback(ogg_packet *packet, int packet_num,
		__a_unused int serial, struct afh_info *afhi, void *private_data)
{
	struct private_vorbis_data *pvd = private_data;

	if (vorbis_synthesis_headerin(&pvd->vi, &pvd->vc, packet) < 0)
		return -E_VORBIS;
	if (packet_num == 0) {
		if (pvd->vi.rate == 0)
			return -E_VORBIS;
		afhi->channels = pvd->vi.channels;
		afhi->frequency = pvd->vi.rate;
		afhi->bitrate = pvd->vi.bitrate_nominal / 1000;
		PARA_DEBUG_LOG("channels: %i, sampling rate: %i, bitrate: %i\n",
			afhi->channels, afhi->frequency, afhi->bitrate);
		return 1;
	}
	if (packet_num == 1)
		return 1; /* we also want to have packet #2 */
	afhi->tags.artist = para_strdup(vorbis_comment_query(&pvd->vc, "artist", 0));
	afhi->tags.title = para_strdup(vorbis_comment_query(&pvd->vc, "title", 0));
	afhi->tags.album = para_strdup(vorbis_comment_query(&pvd->vc, "album", 0));
	afhi->tags.year = para_strdup(vorbis_comment_query(&pvd->vc, "year", 0));
	afhi->tags.comment = para_strdup(vorbis_comment_query(&pvd->vc, "comment", 0));
	return 0;
}

static int ogg_vorbis_get_file_info(char *map, size_t numbytes, __a_unused int fd,
		struct afh_info *afhi)
{
	int ret;
	struct private_vorbis_data pvd;
	struct ogg_afh_callback_info vorbis_callback_info = {
		.packet_callback = vorbis_packet_callback,
		.private_data = &pvd,
	};

	vorbis_info_init(&pvd.vi);
	vorbis_comment_init(&pvd.vc);
	ret = ogg_get_file_info(map, numbytes, afhi, &vorbis_callback_info);
	vorbis_info_clear(&pvd.vi);
	vorbis_comment_clear(&pvd.vc);
	return ret;
}

struct vorbis_get_header_data {
	ogg_stream_state os;
	char *buf;
	size_t len;
};

static void add_ogg_page(ogg_page *og, struct vorbis_get_header_data *vghd)
{
	size_t old_len = vghd->len;
	size_t new_len = vghd->len + og->header_len + og->body_len;
	char *buf = para_realloc(vghd->buf, new_len), *p = buf + old_len;

	memcpy(p, og->header, og->header_len);
	memcpy(p + og->header_len, og->body, og->body_len);
	vghd->buf = buf;
	vghd->len = new_len;
	PARA_DEBUG_LOG("header/body/old/new: %lu/%lu/%zu/%zu\n",
		og->header_len, og->body_len, old_len, new_len);
}

static int vorbis_get_header_callback(ogg_packet *packet, int packet_num,
		int serial, __a_unused struct afh_info *afhi, void *private_data)
{
	int ret;
	struct vorbis_get_header_data *vghd = private_data;
	ogg_page og;
	static unsigned char dummy_packet[] = {
		0x03,
		'v', 'o', 'r', 'b', 'i', 's',
		0x06, 0x00, 0x00, 0x00,
		'd', 'u', 'm', 'm', 'y', '\0',
		0x00, 0x00, 0x00, 0x00, /* no comment :) */
		0xff /* framing bit */
	};

	PARA_DEBUG_LOG("processing ogg packet #%d\n", packet_num);
	if (packet_num > 2)
		return 0;
	if (packet_num == 0) {
		ogg_stream_init(&vghd->os, serial);
		ret = ogg_stream_packetin(&vghd->os, packet);
		if (ret < 0)
			goto out;
		ret = -E_OGG_STREAM_FLUSH;
		if (ogg_stream_flush(&vghd->os, &og) == 0)
			goto out;
		add_ogg_page(&og, vghd);
		return 1;
	}
	if (packet_num == 1) {
		ogg_packet replacement = *packet;
		PARA_INFO_LOG("replacing metadata packet\n");
		replacement.packet = dummy_packet;
		replacement.bytes = sizeof(dummy_packet);
		ret = ogg_stream_packetin(&vghd->os, &replacement);
		if (ret >= 0)
			return 1;
		ret = -E_OGG_PACKET_IN;
		goto out;
	}
	ret = -E_OGG_PACKET_IN;
	if (ogg_stream_packetin(&vghd->os, packet) < 0)
		goto out;
	while (ogg_stream_flush(&vghd->os, &og))
		add_ogg_page(&og, vghd);
	ret = 0;
out:
	ogg_stream_clear(&vghd->os);
	return ret;
}

static void vorbis_get_header(void *map, size_t mapsize, char **buf,
		size_t *len)
{
	int ret;
	struct vorbis_get_header_data vghd = {.len = 0};
	struct ogg_afh_callback_info cb = {
		.packet_callback = vorbis_get_header_callback,
		.private_data = &vghd,
	};

	ret = ogg_get_file_info(map, mapsize, NULL, &cb);
	if (ret < 0)
		goto fail;
	*buf = vghd.buf;
	*len = vghd.len;
	PARA_INFO_LOG("created %zu byte ogg vorbis header\n", *len);
	return;
fail:
	PARA_ERROR_LOG("%s\n", para_strerror(-ret));
}

static int vorbis_make_meta_packet(struct taginfo *tags, ogg_packet *result)
{
	vorbis_comment vc;
	int ret;

	vorbis_comment_init(&vc);
	vorbis_comment_add_tag(&vc, "artist", tags->artist);
	vorbis_comment_add_tag(&vc, "title", tags->title);
	vorbis_comment_add_tag(&vc, "album", tags->album);
	vorbis_comment_add_tag(&vc, "year", tags->year);
	vorbis_comment_add_tag(&vc, "comment", tags->comment);
	ret = vorbis_commentheader_out(&vc, result);
	vorbis_comment_clear(&vc);
	if (ret != 0)
		return -E_VORBIS_COMMENTHEADER;
	return 1;
}

static int vorbis_rewrite_tags(const char *map, size_t mapsize,
		struct taginfo *tags, int output_fd,
		__a_unused const char *filename)
{
	int ret;
	ogg_packet packet;

	ret = vorbis_make_meta_packet(tags, &packet);
	if (ret < 0)
		return ret;
	ret = ogg_rewrite_tags(map, mapsize, output_fd, (char *)packet.packet,
		packet.bytes);
	free(packet.packet);
	return ret;
}

static const char * const ogg_suffixes[] = {"ogg", NULL};

/**
 * The init function of the ogg vorbis audio format handler.
 *
 * \param afh Pointer to the struct to initialize.
 */
void ogg_init(struct audio_format_handler *afh)
{
	afh->get_file_info = ogg_vorbis_get_file_info;
	afh->get_header = vorbis_get_header;
	afh->suffixes = ogg_suffixes;
	afh->rewrite_tags = vorbis_rewrite_tags;
}
